from datasette.database import QueryInterrupted
from datasette.facets import Facet
from datasette.utils import (
    escape_sqlite,
    path_with_added_args,
    path_with_removed_args,
)

class YearFacet(Facet):
    type = "year"

    async def suggest(self):
        return []

    async def facet_results(self):
        facet_results = []
        facets_timed_out = []
        args = dict(self.get_querystring_pairs())
        facet_size = self.get_facet_size()
        for source_and_config in self.get_configs():
            config = source_and_config["config"]
            source = source_and_config["source"]
            column = config.get("column") or config["simple"]
            # TODO: does this query break if inner sql produces value or count columns?
            facet_sql = """
                select substr(date({col}), 1, 4) as value, count(*) as count from (
                    {sql}
                )
                where date({col}) is not null
                group by substr(date({col}), 1, 4) order by count desc, value limit {limit}
            """.format(
                col=escape_sqlite(column), sql=self.sql, limit=facet_size + 1
            )
            try:
                facet_rows_results = await self.ds.execute(
                    self.database,
                    facet_sql,
                    self.params,
                    truncate=False,
                    custom_time_limit=self.ds.setting("facet_time_limit_ms"),
                )
                facet_results_values = []
                facet_results.append(
                    {
                        "name": column,
                        "type": self.type,
                        "results": facet_results_values,
                        "hideable": source != "metadata",
                        "toggle_url": path_with_removed_args(
                            self.request, {"_facet_year": column}
                        ),
                        "truncated": len(facet_rows_results) > facet_size,
                    }
                )
                facet_rows = facet_rows_results.rows[:facet_size]
                for row in facet_rows:
                    selected = str(args.get(f"{column}__like")) == str(row["value"]) + '%'
                    if selected:
                        toggle_path = path_with_removed_args(
                            self.request, {f"{column}__like": str(row["value"]) + '%'}
                        )
                    else:
                        toggle_path = path_with_added_args(
                            self.request, {f"{column}__like": row["value"] + '%'}
                        )
                    facet_results_values.append(
                        {
                            "value": row["value"],
                            "label": row["value"],
                            "count": row["count"],
                            "toggle_url": self.ds.absolute_url(
                                self.request, toggle_path
                            ),
                            "selected": selected,
                        }
                    )
            except QueryInterrupted:
                facets_timed_out.append(column)

        return facet_results, facets_timed_out

class YearMonthFacet(Facet):
    type = "year_month"

    async def suggest(self):
        return []

    async def facet_results(self):
        facet_results = []
        facets_timed_out = []
        args = dict(self.get_querystring_pairs())
        facet_size = self.get_facet_size()
        for source_and_config in self.get_configs():
            config = source_and_config["config"]
            source = source_and_config["source"]
            column = config.get("column") or config["simple"]
            # TODO: does this query break if inner sql produces value or count columns?
            facet_sql = """
                select substr(date({col}), 1, 7) as value, count(*) as count from (
                    {sql}
                )
                where date({col}) is not null
                group by substr(date({col}), 1, 7) order by count desc, value limit {limit}
            """.format(
                col=escape_sqlite(column), sql=self.sql, limit=facet_size + 1
            )
            try:
                facet_rows_results = await self.ds.execute(
                    self.database,
                    facet_sql,
                    self.params,
                    truncate=False,
                    custom_time_limit=self.ds.setting("facet_time_limit_ms"),
                )
                facet_results_values = []
                facet_results.append(
                    {
                        "name": column,
                        "type": self.type,
                        "results": facet_results_values,
                        "hideable": source != "metadata",
                        "toggle_url": path_with_removed_args(
                            self.request, {"_facet_year_month": column}
                        ),
                        "truncated": len(facet_rows_results) > facet_size,
                    }
                )
                facet_rows = facet_rows_results.rows[:facet_size]
                for row in facet_rows:
                    selected = str(args.get(f"{column}__like")) == str(row["value"]) + '%'
                    if selected:
                        toggle_path = path_with_removed_args(
                            self.request, {f"{column}__like": str(row["value"]) + '%'}
                        )
                    else:
                        toggle_path = path_with_added_args(
                            self.request, {f"{column}__like": row["value"] + '%'}
                        )
                    facet_results_values.append(
                        {
                            "value": row["value"],
                            "label": row["value"],
                            "count": row["count"],
                            "toggle_url": self.ds.absolute_url(
                                self.request, toggle_path
                            ),
                            "selected": selected,
                        }
                    )
            except QueryInterrupted:
                facets_timed_out.append(column)

        return facet_results, facets_timed_out

class StatsFacet(Facet):
    type = "stats"

    async def suggest(self):
        return []

    async def facet_results(self):
        facet_results = []
        facets_timed_out = []
        args = dict(self.get_querystring_pairs())
        facet_size = self.get_facet_size()
        for source_and_config in self.get_configs():
            config = source_and_config["config"]
            source = source_and_config["source"]
            column = config.get("column") or config["simple"]
            # TODO: does this query break if inner sql produces value or count columns?
            facet_sql = """
with
xs as ({sql}),
with_ids as (select row_number() over () as id, {col} as value from xs),
ntiles as (select id, value, ntile(100) over (order by value) percentile from with_ids)
select percentile, min(value), max(value) from ntiles
group by 1
having percentile in (1, 50, 90, 95, 99, 100)
order by 1
            """.format(
                col=escape_sqlite(column), sql=self.sql, limit=facet_size + 1
            )
            try:
                facet_rows_results = await self.ds.execute(
                    self.database,
                    facet_sql,
                    self.params,
                    truncate=False,
                    custom_time_limit=self.ds.setting("facet_time_limit_ms"),
                )
                facet_results_values = []
                facet_results.append(
                    {
                        "name": column,
                        "type": self.type,
                        "results": facet_results_values,
                        "hideable": source != "metadata",
                        "toggle_url": path_with_removed_args(
                            self.request, {"_facet_stats": column}
                        ),
                        "truncated": len(facet_rows_results) > facet_size,
                    }
                )
                facet_rows = facet_rows_results.rows[:facet_size]

                percentile_groups = {}
                for row in facet_rows:
                    percentile_groups['{}_min'.format(row[0])] = row[1]
                    percentile_groups['{}_max'.format(row[0])] = row[2]

                facet_rows = [
                    ('max', percentile_groups['100_max']),
                    ('p99', percentile_groups['99_max']),
                    ('p95', percentile_groups['95_max']),
                    ('p90', percentile_groups['90_max']),
                    ('median', percentile_groups['50_min']),
                    ('min', percentile_groups['1_min']),
                ]

                for value, count in facet_rows:
                    facet_results_values.append(
                        {
                            "value": value,
                            "label": value,
                            "count": count,
                            "toggle_url": '#',
                            "selected": False,
                        }
                    )
            except QueryInterrupted:
                facets_timed_out.append(column)

        return facet_results, facets_timed_out



import sqlite3
import time
import json
import asyncio
from .column_stats_schema import ensure_schema, DUX_COLUMN_STATS, DUX_COLUMN_STATS_VALUES
import traceback
import sys

# A table is indexable if:
# - not a virtual table (eg fts_posts)
# - not a suffix of a virtual table (eg, fts_posts_idx)
# - not a dux_ table
async def indexable_tables(db):
    return [row[0] for row in list(await db.execute("select name from sqlite_master sm where type = 'table' and not sql like 'CREATE VIRTUAL TABLE%' and not name like 'dux_%' and not exists(select 1 from sqlite_master sm2 where type = 'table' and sql like 'CREATE VIRTUAL TABLE%' and sm.name like sm2.name || '_%')"))]

async def indexable_columns(db, table):
    columns = list(await db.execute('select name from pragma_table_info(?)', [table]))
    return [row[0] for row in columns]

async def compute_dux_column_stats(ds):
    t = time.time()

    dbs_to_watch = []
    for db_name, db in ds.databases.items():
        if db.is_memory or not db.is_mutable:
            continue

        if hasattr(db, 'engine') and db.engine == 'duckdb':
            continue

        await ensure_schema(db)
        dbs_to_watch.append(db_name)

    ds._dux_dbs_to_index = dbs_to_watch

def start_column_stats_indexer(datasette):
    if not datasette._dux_dbs_to_index:
        return

    loop = asyncio.get_running_loop()
    # We need to keep a reference to this task to prevent GC?
    # https://news.ycombinator.com/item?id=34754276
    datasette._dux_indexer_task = loop.create_task(index_loop(datasette, datasette._dux_dbs_to_index))

async def ensure_id(db, identifier):
    existing = await db.execute('SELECT id FROM dux_ids WHERE name = ?', [identifier])

    if existing:
        return list(existing)[0][0]

    if not existing:
        await db.execute_write('INSERT INTO dux_ids(name) VALUES (?)', [identifier])

        return await ensure_id(db, identifier)

async def ensure_ids(db, ids, needles): 
    missing = []
    for needle in list(set(needles)):
        if not needle in ids:
            missing.append(needle)

    if not missing:
        return

    for name in missing:
        ids[name] = await ensure_id(db, name)

async def ensure_empty_rows_for_db(db):
    known_tables = {}
    tables = await indexable_tables(db)

    ids = {}
    all_ids = await db.execute('SELECT id, name FROM dux_ids')
    for row in all_ids:
        ids[row['name']] = row['id']

    known_ops_raw = list(await db.execute('SELECT table_id, column_id FROM dux_column_stats_ops'))
    known_ops = {}
    for table_id, column_id in known_ops_raw:
        known_ops[(table_id, column_id)] = True

    known_stats_raw = list(await db.execute('SELECT table_id, column_id FROM dux_column_stats'))
    known_stats = {}
    for table_id, column_id in known_stats_raw:
        known_stats[(table_id, column_id)] = True

    await ensure_ids(db, ids, tables)
    for table in tables:
        table_id = ids[table]
        columns = await indexable_columns(db, table)
        await ensure_ids(db, ids, columns)
        column_ids = []
        known_tables[table_id] = column_ids

        column_infos = list(await db.execute('select name, type, "notnull", pk from pragma_table_info(?)', [table]))
        column_info_by_name = {}
        for column_info in column_infos:
            column_info_by_name[column_info['name']] = column_info

        for column in columns:
            column_id = ids[column]
            column_ids.append(column_id)

            column_info = column_info_by_name.get(column, None)
            if not column_info:
                continue

            column, type, notnull, pk = column_info
            nullable = notnull == 0

            if not (table_id, column_id) in known_ops:
                await db.execute_write(
                    'INSERT INTO dux_column_stats_ops(table_id, column_id, pending) SELECT ?, ?, 1 WHERE NOT EXISTS(SELECT * FROM dux_column_stats_ops WHERE table_id = ? AND column_id = ?)',
                   [table_id, column_id, table_id, column_id]
                )

            if not (table_id, column_id) in known_stats:
                await db.execute_write(
                    'INSERT INTO dux_column_stats(table_id, column_id, type, nullable, pk) SELECT ?, ?, ?, ?, ? WHERE NOT EXISTS(SELECT * FROM dux_column_stats WHERE table_id = ? AND column_id = ?)',
                   [table_id, column_id, type, nullable, pk, table_id, column_id]
                )

    # Remove entries for tables that no longer exist
    # Eventually we'll want to be able to do this online, but for now you have to restart.
    known_table_ids = ', '.join([str(x) for x in known_tables.keys()])
    if known_table_ids:
        await db.execute_write('DELETE FROM dux_column_stats_ops WHERE NOT table_id IN ({})'.format(known_table_ids))
        await db.execute_write('DELETE FROM dux_column_stats_values WHERE NOT table_id IN ({})'.format(known_table_ids))
        await db.execute_write('DELETE FROM dux_column_stats WHERE NOT table_id IN ({})'.format(known_table_ids))

    # TODO: delete entries that are in dux_column_stats but no longer in columns
    # TODO: delete entries that are in dux_column_stats_values but no longer in columns


async def index_loop(ds, dbs_to_watch):
    try:
        # Ensure that there are default empty rows in dux_column_stats_ops, dux_column_stats for every column.
        t = time.time()
        for db_name in dbs_to_watch:
            db = ds.databases[db_name]
            await ensure_empty_rows_for_db(db)

        print('datasette-ui-extras: stats schemas ensured in {} s'.format(time.time() - t))

        while True:
            did_work = False
            for db_name in dbs_to_watch:
                await asyncio.sleep(0)

                db = ds.databases[db_name]
                did_work = await index_db(db) or did_work

            if not did_work:
                await asyncio.sleep(1)

    except Exception as e:
       print('datasette-ui-extras: error during index_loop; exiting', file=sys.stderr)
       traceback.print_exc()
       sys.exit(1)

async def index_db(db):
    chunk_size = 1000

    # Look for a column that needs some progress.
    next_row = list(await db.execute('select table_id, column_id, last_key, (select name from dux_ids ids where ids.id = ops.table_id) AS table_name, (select name from dux_ids ids where ids.id = ops.column_id) AS column_name, updated_at from dux_column_stats_ops ops where pending order by updated_at asc limit 1'))

    if not next_row:
        return

    next_row = next_row[0]
    # print([x for x in next_row])
    table_id, column_id, last_key, table_name, column_name, updated_at = next_row

    pks = list(await db.execute('select name from pragma_table_info(?) where pk', [table_name]))
    pks = [row[0] for row in pks]
    if not pks:
        pks = ['rowid']

    key_expr = 'json_object({})'.format(
        ', '.join(["'{}'".format(pk) + ', ' + '"{}"'.format(pk) for pk in pks])
    )
    order_by = ', '.join(['"{}"'.format(pk) for pk in pks])
    last_key = json.loads(last_key)
    where = ''
    where_args = []
    if last_key:
        where = 'WHERE ({pks}) > ({last_keys})'.format(
            pks=', '.join(['"{}"'.format(pk) for pk in pks]),
            last_keys = ', '.join(['?' for pk in pks]),
        )
        where_args = [last_key[pk] for pk in pks]

    def make_base_sql(include_value=True, limit=chunk_size):
        value = ""

        if include_value:
            value = ", {value} AS value ".format(value = '"{}"'.format(column_name))
        base_sql = 'SELECT {key} AS key {value} FROM {table_name} {where} ORDER BY {order_by} LIMIT {limit}'.format(
            key=key_expr,
            value=value,
            table_name='"{}"'.format(table_name),
            where=where,
            order_by=order_by,
            limit=limit
        )
        return base_sql

    sql_including_values = make_base_sql(True)

    await update_summary_stats(db, table_id, column_id, sql_including_values, where_args)
    await update_distincts(db, table_id, column_id, sql_including_values, where_args)
    await update_json_distincts(db, table_id, column_id, sql_including_values, where_args)

    # Fetch and update distinct json array values

    # Determine what new value for last_key should be and update the ops table.
    next_key = list(await db.execute('{} OFFSET {}'.format(make_base_sql(False, 1), chunk_size - 1), where_args))


    if next_key:
        next_key = next_key[0][0]
    else:
        next_key = '{}'

    await db.execute_write(
        "UPDATE dux_column_stats_ops SET last_key = ?, pending = ?, updated_at = strftime('%Y-%m-%d %H:%M:%f') || 'Z' WHERE table_id = ? AND column_id = ?",
        [next_key, 0 if next_key == '{}' else 1, table_id, column_id]
    )

    return True

async def update_distincts(db, table_id, column_id, sql, where_args):
    # Fetch and update distinct values
    distincts_sql = '''
WITH xs AS ({}),
distincts AS (
SELECT
  substr(lower(value), 1, 20) AS value,
  substr(md5(value), 1, 8) AS hash,
  count(*) as count,
  json_group_array(json(key)) AS keys
FROM xs
WHERE typeof(value) == 'text' AND length(value) <= 100 and not (value >= '1800-01-01' and value <= '9999-12-31') and not value like 'http://%' and not value like 'https://%' and cast(cast(value as integer) as text) != value and not (value like '[%' and json_valid(value))
GROUP BY 1, 2
ORDER BY 3 DESC
)
SELECT value, hash, count, (select json_group_array(value) from (select value from json_each(keys) limit 10)) AS keys
FROM distincts
'''.format(sql)

    def read_fn(conn):
        return conn.execute(distincts_sql, where_args).fetchall()
    # NB: use execute_fn to avoid time limits
    distincts = await db.execute_fn(read_fn)

    def fn(conn):
        for value, hash, count, pks in distincts:
            pks = json.loads(pks)
            exists = conn.execute('SELECT pks FROM dux_column_stats_values WHERE table_id = ? and column_id = ? AND value = ? AND hash = ?', [table_id, column_id, value, hash]).fetchone()

            new_pks = []
            if exists:
                new_pks = json.loads(exists[0])
            else:
                conn.execute("INSERT INTO dux_column_stats_values(table_id, column_id, value, hash, count, pks) VALUES (?, ?, ?, ?, 0, '[]')", [table_id, column_id, value, hash]).fetchall()

            new_pks = [pk for pk in new_pks if not pk in pks]
            new_pks = new_pks[0:10 - len(pks)]
            new_pks = new_pks + pks

            conn.execute("UPDATE dux_column_stats_values SET count = count + ?, pks = ? WHERE table_id = ? AND column_id = ? AND value = ? AND hash = ?", [count, json.dumps(new_pks), table_id, column_id, value, hash]).fetchall()

    await db.execute_write_fn(fn)

async def update_json_distincts(db, table_id, column_id, sql, where_args):
    # Fetch and update distinct values from JSON string arrays
    json_distincts_sql = '''
WITH raw_xs AS ({}),
array_xs AS (SELECT key, value FROM raw_xs WHERE json_valid(value) AND json_type(value) = 'array'),
xs AS (SELECT array_xs.key, js_value.value FROM array_xs, json_each(array_xs.value) js_value),
distincts AS (
SELECT
  substr(lower(value), 1, 20) AS value,
  substr(md5(value), 1, 8) AS hash,
  count(*) as count,
  json_group_array(json(key)) AS keys
FROM xs
WHERE typeof(value) == 'text' AND length(value) <= 100 and not (value >= '1800-01-01' and value <= '9999-12-31') and not value like 'http://%' and not value like 'https://%' and cast(cast(value as integer) as text) != value and not (value like '[%' and json_valid(value))
GROUP BY 1, 2
ORDER BY 3 DESC
)
SELECT value, hash, count, (select json_group_array(value) from (select value from json_each(keys) limit 10)) AS keys
FROM distincts
'''.format(sql)

    def read_fn(conn):
        return conn.execute(json_distincts_sql, where_args).fetchall()

    # NB: use excute_fn to avoid time limits
    distincts = await db.execute_fn(read_fn)

    def fn(conn):
        for value, hash, count, pks in distincts:
            pks = json.loads(pks)
            exists = conn.execute('SELECT pks FROM dux_column_stats_values WHERE table_id = ? and column_id = ? AND value = ? AND hash = ?', [table_id, column_id, value, hash]).fetchone()

            new_pks = []
            if exists:
                new_pks = json.loads(exists[0])
            else:
                conn.execute("INSERT INTO dux_column_stats_values(table_id, column_id, value, hash, count, pks) VALUES (?, ?, ?, ?, 0, '[]')", [table_id, column_id, value, hash]).fetchall()

            new_pks = [pk for pk in new_pks if not pk in pks]
            new_pks = new_pks[0:10 - len(pks)]
            new_pks = new_pks + pks

            conn.execute("UPDATE dux_column_stats_values SET count = count + ?, pks = ? WHERE table_id = ? AND column_id = ? AND value = ? AND hash = ?", [count, json.dumps(new_pks), table_id, column_id, value, hash]).fetchall()

    await db.execute_write_fn(fn)


async def update_summary_stats(db, table_id, column_id, sql, where_args):
    # Fetch and update summary stats
    summary_stats = list(await db.execute('''
WITH xs AS ({})
SELECT
  min(case when typeof(value) == 'blob' or (typeof(value) == 'text' and length(value) > 100) then null else value end) as min,
  max(case when typeof(value) == 'blob' or (typeof(value) == 'text' and length(value) > 100) then null else value end) as max,
  count(*) as count,
  count(*) filter (where typeof(value) == 'null') as nulls,
  count(*) filter (where typeof(value) == 'integer') as integers,
  count(*) filter (where typeof(value) == 'real') as reals,
  count(*) filter (where typeof(value) == 'text') as texts,
  count(*) filter (where typeof(value) == 'blob') as blobs,
  coalesce(sum(case when json_valid(value) then json_type(value) = 'text' else 0 end), 0) as json_strings,
  coalesce(sum(case when json_valid(value) then json_type(value) = 'array' else 0 end), 0) as json_arrays,
  coalesce(sum(case when json_valid(value) then json_type(value) = 'object' else 0 end), 0) as json_objects,
  min(length(value)) filter (where typeof(value) == 'text') as texts_min_length,
  max(length(value)) filter (where typeof(value) == 'text') as texts_max_length,
  coalesce(sum(case when typeof(value) == 'text' and value like '% %' then 1 else 0 end), 0) as texts_whitespace,
  coalesce(sum(case when typeof(value) == 'text' and value like '%' || x'0a' || '%' then 1 else 0 end), 0) as texts_newline,
  min(length(value)) filter (where typeof(value) == 'blob') as blobs_min_length,
  max(length(value)) filter (where typeof(value) == 'blob') as blobs_max_length
FROM xs
'''.format(sql), where_args))

    # texts_whitespace, texts_newline
    if summary_stats:
        summary_stats = summary_stats[0]
        #print('jsons={}'.format(summary_stats['jsons']))
        await db.execute_write('''
    UPDATE dux_column_stats SET
        count = count + :count,
        max = CASE WHEN max IS NULL THEN :max ELSE max(max, :max) END,
        min = CASE WHEN min IS NULL THEN :min ELSE min(min, :min) END,
        nulls = nulls + :nulls,
        integers = integers + :integers,
        reals = reals + :reals,
        texts = texts + :texts,
        blobs = blobs + :blobs,
        json_strings = json_strings + :json_strings,
        json_arrays = json_arrays + :json_arrays,
        json_objects = json_objects + :json_objects,
        texts_min_length = CASE WHEN texts_min_length IS NULL THEN :texts_min_length ELSE min(texts_min_length, :texts_min_length) END,
        texts_max_length = CASE WHEN texts_max_length IS NULL THEN :texts_max_length ELSE max(texts_max_length, :texts_max_length) END,
        blobs_min_length = CASE WHEN blobs_min_length IS NULL THEN :blobs_min_length ELSE min(blobs_min_length, :blobs_min_length) END,
        blobs_max_length = CASE WHEN blobs_max_length IS NULL THEN :blobs_max_length ELSE max(blobs_max_length, :blobs_max_length) END,
        texts_whitespace = texts_whitespace + :texts_whitespace,
        texts_newline = texts_newline + :texts_newline
    WHERE table_id = :table_id AND column_id = :column_id
    ''', {
        'table_id': table_id,
        'column_id': column_id,
        'count': summary_stats['count'],
        'min': summary_stats['min'],
        'max': summary_stats['max'],
        'nulls': summary_stats['nulls'],
        'integers': summary_stats['integers'],
        'reals': summary_stats['reals'],
        'texts': summary_stats['texts'],
        'blobs': summary_stats['blobs'],
        'json_strings': summary_stats['json_strings'],
        'json_arrays': summary_stats['json_arrays'],
        'json_objects': summary_stats['json_objects'],
        'texts_min_length': summary_stats['texts_min_length'],
        'texts_max_length': summary_stats['texts_max_length'],
        'blobs_min_length': summary_stats['blobs_min_length'],
        'blobs_max_length': summary_stats['blobs_max_length'],
        'texts_whitespace': summary_stats['texts_whitespace'],
        'texts_newline': summary_stats['texts_newline'],
    })


def autosuggest_column(conn, table, column, q):
    # Consider max 100 things before sorting - gives generally good results, and is defensive
    # against someone autocompleting the empty string
    raw_rows = conn.execute(
        ('WITH xs AS (SELECT value, hash, count, pks FROM {} WHERE "table_id" = (SELECT id FROM dux_ids WHERE name = ?) AND "column_id" = (SELECT id FROM dux_ids WHERE name = ?) AND value >= ? AND value < ? || ' + "x'ffffffff'" + ' LIMIT 100) SELECT * FROM xs ORDER BY count DESC LIMIT 10').format(DUX_COLUMN_STATS_VALUES),
        [table, column, q.lower(), q.lower()]
    ).fetchall()

    # The value stored is just for indexing -- we need to fetch the actual value
    # from the underlying table, by looking up the entries in `pks`.
    rows = []
    for value, hash, count, pks in raw_rows:
        pks = json.loads(pks)
        if not pks:
            continue

        pk = pks[0]
        keys = list(pk.items())
        rv = conn.execute(
            '''
WITH xs as (SELECT "{column}" AS value FROM "{table}" WHERE ({keys}) = ({key_bindings})),
choices AS (SELECT json.value, substr(md5(json.value), 1, 8) as hash FROM xs, json_each(CASE WHEN json_valid(xs.value) AND json_type(xs.value) = 'array' THEN xs.value ELSE json_array(xs.value) END) AS json)
SELECT value FROM choices WHERE hash = ?
            '''.format(
                column = column,
                table = table,
                keys = ', '.join([k[0] for k in keys]),
                key_bindings = ', '.join(['?' for k in keys])
            ),
            [k[1] for k in keys] + [hash]
        ).fetchone()

        if not rv:
            continue

        rows.append({
            'value': rv[0],
            'count': count,
            'pks': pks
        })


    return rows



import pytest

from datasette_ui_extras.schema_utils import get_column_choices_from_check_constraints

def test_column_choices_from_check_constraints():
    schema = '''CREATE TABLE q(
a check (a in ('foo', 'bar')),
b,
c check (c > 5),
check (b in (1, 2)))'''

    constraints = get_column_choices_from_check_constraints(schema)
    assert constraints == {'a': ['foo', 'bar'], 'b': [1, 2]}

    assert {} == get_column_choices_from_check_constraints('invalid sql')



from datasette.utils.asgi import Forbidden
from .utils import row_edit_params

def enable_yolo_edit_row_pages():
    # We also enable a new template: edit-row.
    from datasette.views.row import RowView

    # We can't capture the original RowView.data function outside of this fn
    # due to a circular import issue, so do it here.
    #
    # In tests, this can lead to patching it twice, so remember if we've patched it.
    if not hasattr(RowView, 'dux_patched'):
        RowView.dux_patched = True
        original_RowView_data = RowView.data
        async def patched_RowView_data(self, request, default_labels=False):
            data, template_data, templates = await original_RowView_data(self, request, default_labels)


            resolved = await self.ds.resolve_row(request)
            database = resolved.db.name
            table = resolved.table

            edit_mode = await row_edit_params(self.ds, request, database, table)

            if edit_mode:
                templates = tuple(['edit-row' + x[3:] for x in templates])
            return (data, template_data, templates)

        RowView.data = patched_RowView_data



import sqlite3
import pytest

def test_compute_column_stats(tmp_path):
    # TODO: re-write this test to be compatible with the new indexing system
    if False:
        conn = sqlite3.connect(db_name)
        with conn:
            conn.execute("CREATE TABLE data(id integer primary key, title text, json text, age integer, blobby blob, ratio real)")
            conn.execute("INSERT INTO data(id, title, json, age, ratio) VALUES (1, 'title1', '[\"foo\"]', 23, 0.43)").fetchall()
            conn.execute("INSERT INTO data(id, title, json) VALUES (2, 'title2', 'not json')").fetchall()

        conn.row_factory = sqlite3.Row
        limit = 1000
        distinct_limit = 100
        rv = compute_column_stats(conn, 'data', 'json', limit, distinct_limit)



from datasette.database import Database
from sqlite_utils.db import View, Table, validate_column_names, jsonify_if_needed
from sqlglot import parse_one, exp
from .utils import get_view_info

# Adds two features:
# - can navigate to the row page for a view row, eg /db/viewname/1
# - can use the JSON API to do an update of a view row (...assumes there's an INSTEAD OF trigger)

original_primary_keys = Database.primary_keys

async def patched_primary_keys(self, table):
    # Is it a view?
    # Get its definition
    # Parse it with sqlglot
    # Is it a select that queries exactly one table?
    is_view = list(await self.execute("select sql from sqlite_master where type = 'view' and name = ?", [table]))

    if not is_view:
        return await original_primary_keys(self, table)

    def fn(conn):
        return get_view_info(conn, table)

    view_info = await self.execute_fn(fn)

    if not view_info:
        return []

    return view_info['pks']

original_get_all_foreign_keys = Database.get_all_foreign_keys
async def patched_get_all_foreign_keys(self):
    rv = await original_get_all_foreign_keys(self)

    names = await self.execute("select name from sqlite_master where type = 'view'")
    for name, in names:
        rv[name] = {'incoming': [], 'outgoing': []}

    return rv

def get_view_primary_keys(conn, name):
    info = get_view_info(conn, name)
    if 'base_table' in info:
        return info['pks']

    return []


class UpdateableView(View):
    def __init__(self, underlying):
        self.underlying = underlying

    @property
    def pks(self):
        "Primary key columns for this view."

        pks = get_view_primary_keys(self.db.conn, self.name)

        if pks:
            return pks

        return ['rowid']

    def __getattr__(self, name):
        return getattr(self.underlying, name)

    def update(
        self,
        pk_values,
        updates = None,
        alter = False,
        conversions = None,
    ):
        """
        Execute a SQL ``UPDATE`` against the specified row.

        See :ref:`python_api_update`.

        :param pk_values: The primary key of an individual record - can be a tuple if the
          table has a compound primary key.
        :param updates: A dictionary mapping columns to their updated values.
        :param alter: Set to ``True`` to add any missing columns.
        :param conversions: Optional dictionary of SQL functions to apply during the update, for example
          ``{"mycolumn": "upper(?)"}``.
        """
        updates = updates or {}
        conversions = conversions or {}
        if not isinstance(pk_values, (list, tuple)):
            pk_values = [pk_values]
        # Soundness check that the record exists (raises error if not):
        self.get(pk_values)
        if not updates:
            return self
        args = []
        sets = []
        wheres = []
        pks = self.pks
        validate_column_names(updates.keys())
        for key, value in updates.items():
            sets.append("[{}] = {}".format(key, conversions.get(key, "?")))
            args.append(jsonify_if_needed(value))
        wheres = ["[{}] = ?".format(pk_name) for pk_name in pks]
        args.extend(pk_values)
        sql = "update [{table}] set {sets} where {wheres}".format(
            table=self.name, sets=", ".join(sets), wheres=" and ".join(wheres)
        )
        #print('running update: sql={} args={}'.format(sql, args))
        with self.db.conn:
            rowcount = self.db.execute(sql, args).rowcount

            # NOTE: Don't check rowcount, updates on updatable views don't have a rowcount.
            # assert rowcount == 1
        self.last_pk = pk_values[0] if len(pks) == 1 else pk_values
        return self


def thunk_update(
    self,
    pk_values,
    updates = None,
    alter = False,
    conversions = None,
):

    if alter:
        raise Exception('updates on view must have alter=False')

    if conversions:
        raise Exception('updates on view must have conversions=None')

    uv = UpdateableView(self)
    return uv.update(pk_values, updates, alter=False, conversions=None)

def enable_yolo_view_row_pages():
    Database.primary_keys = patched_primary_keys
    Database.get_all_foreign_keys = patched_get_all_foreign_keys

    # Enable updating views in the JSON API. I didn't want to snapshot the code,
    # nor generally .get and .pks on views, so I do some shenanigans here to
    # add an .update method that gets hijacked by UpdateableView.
    #
    # I ended up having to snapshot the code, because the existing stuff did
    # an assert on the rowcount -- which is not correct for INSTEAD OF triggers.
    UpdateableView.get = Table.get
    View.update = thunk_update



from datasette.app import Datasette
import pytest
import json
import sqlite3


@pytest.mark.asyncio
async def test_plugin_is_installed():
    datasette = Datasette(memory=True)
    response = await datasette.client.get("/-/plugins.json")
    assert response.status_code == 200
    installed_plugins = {p["name"] for p in response.json()}
    assert "datasette-ui-extras" in installed_plugins

@pytest.mark.asyncio
async def test_updatable_view(tmp_path):
    db_name = tmp_path / "db.sqlite"
    conn = sqlite3.connect(db_name)
    with conn:
        conn.execute("CREATE TABLE data(id integer primary key, title text, other text)")
        conn.execute("INSERT INTO data(id, title) VALUES (1, 'title1')").fetchall()
        conn.execute("INSERT INTO data(id, title) VALUES (2, 'title2')").fetchall()
        conn.execute("CREATE VIEW view AS SELECT id, 'some jazz' as instructions, title, other as uneditable FROM data")
        conn.execute("""
CREATE TRIGGER update_underlying_fields INSTEAD OF UPDATE OF title ON view
BEGIN
   UPDATE data SET title = new.title WHERE id = old.id;
END;
""")
    conn.close()

    datasette = Datasette(
        memory=True,
        plugins_dir='./tests/plugins/',
        files=[db_name]
    )

    response = await datasette.client.get("/db.json")
    assert response.status_code == 200

    response = await datasette.client.get("/db/data.json")
    assert response.status_code == 200

    response = await datasette.client.get("/db/data/1.json?_shape=array")
    assert response.status_code == 200
    assert response.json() == [{'id': 1, 'title': 'title1', 'other': None}]

    response = await datasette.client.get("/db/view/1.json?_shape=array")
    assert response.status_code == 200
    assert response.json() == [{'id': 1, 'instructions': 'some jazz', 'uneditable': None, 'title': 'title1'}]

    response = await datasette.client.post(
        "/db/view/1/-/update",
        headers={
            'content-type': 'application/json',
        },
        content=json.dumps({
            "update": {
                "title": "new title",
            }
        })
    )
    assert response.status_code == 200

    response = await datasette.client.get("/db/view/1.json?_shape=array&_col=title")
    assert response.status_code == 200
    assert response.json() == [{'id': 1, 'instructions': 'some jazz', 'uneditable': None, 'title': 'new title'}]

    response = await datasette.client.get("/db/view/1")
    assert response.status_code == 200


# Test cases to add
# - ?_edit_dux=1 when you lack update-row should 403
# - ?_edit_dux=1 when you have update-row should 200



import json
import os
from functools import wraps
import hashlib
import datasette
from datasette import Response
from datasette.utils import await_me_maybe
import asyncio
import sqlite_sqlean
import markupsafe
from urllib.parse import parse_qs
from .hookspecs import hookimpl
from .facets import enable_yolo_facets, facets_extra_body_script
from .filters import enable_yolo_arraycontains_filter, enable_yolo_exact_filter, yolo_filters_from_request
from .plugin import pm
from .edit_controls import render_cell_edit_control
from .new_facets import StatsFacet, YearFacet, YearMonthFacet
from .view_row_pages import enable_yolo_view_row_pages
from .edit_row_pages import enable_yolo_edit_row_pages
from .utils import is_row_page
from .yolo_command import yolo_command
from .column_stats_schema import DUX_IDS, DUX_COLUMN_STATS, DUX_COLUMN_STATS_OPS, DUX_COLUMN_STATS_VALUES
from .column_stats import compute_dux_column_stats, autosuggest_column, start_column_stats_indexer

PLUGIN = 'datasette-ui-extras'

css_files = [
    "app.css",
    "hide-export.css",
    "hide-table-definition.css",
    "sticky-table-headers.css",
    "lazy-facets.css",
    "hide-filters.css",
    "layout-row-page.css",
    "compact-cogs.css",
    "mobile-column-menu.css",
    "edit-row/",
]

js_files = [
    'hide-filters.js',
    "sticky-table-headers.js",
    "focus-search-box.js",
    'lazy-facets.js',
    'layout-row-page.js',
    "mobile-column-menu.js",
    "edit-row/",
]

def fingerprint(files, ext):
    def concatenate():
        rv = []
        for fname in files:

            fpaths = [os.path.abspath(os.path.join(__file__, '..', 'static', fname))]

            if fname.endswith('/'):
                fpaths = []

                for fname in sorted(os.scandir(os.path.join(os.path.abspath(os.path.join(__file__, '..', 'static')), fname[:-1])), key=lambda f: f.path):
                    if fname.path.endswith('.' + ext):
                        fpaths.append(os.path.abspath(fname.path))


            for fpath in fpaths:
                rv.append('/* {} */'.format(fpath))
                f = open(fpath, 'r')
                contents = f.read()
                f.close()
                rv.append(contents)

        return '\n\n'.join(rv)

    hashcode = hashlib.sha256(concatenate().encode('utf-8')).hexdigest()[0:8]

    # TODO: how can we distinguish prod vs dev so we can serve a fingerprinted,
    #       long-lived cached file in prod, but have live reload in dev?
    #path = '/-/{}/{}.{}'.format(PLUGIN, hashcode, ext)
    path = '/-/{}/{}.{}'.format(PLUGIN, ext, ext)

    return path, concatenate

css_path, css_contents = fingerprint(css_files, 'css')
js_path, js_contents = fingerprint(js_files, 'js')

# Not fully fleshed out: datasette-ui-extras consists of a bunch of "extras".
# Each extra has one or more of: custom CSS, custom JS, custom Python.
#
# Try to develop them as though they're standalone things, so they can be
# easily turned on/off, or upstreamed into Datasette.
@datasette.hookimpl
def extra_css_urls(datasette):
    return [
        css_path,
        'https://cdnjs.cloudflare.com/ajax/libs/awesomplete/1.1.5/awesomplete.min.css',
        'https://cdn.jsdelivr.net/npm/air-datepicker@3.3.5/air-datepicker.min.css'
    ]

@datasette.hookimpl()
def extra_js_urls(datasette):
    return [
        js_path,
        'https://cdnjs.cloudflare.com/ajax/libs/awesomplete/1.1.5/awesomplete.min.js',
        'https://cdn.jsdelivr.net/npm/air-datepicker@3.3.5/air-datepicker.min.js'
    ]

@datasette.hookimpl
def render_cell(datasette, database, table, column, value):
    async def inner():
        task = asyncio.current_task()
        request = None if not hasattr(task, '_duxui_request') else task._duxui_request

        # Are we on the row page?
        # This feels like quite a large hack!
        if is_row_page(request):
            db = datasette.databases[database]

            # Is this column an fkey?
            fkeys = list(await db.execute('SELECT "table", "to" FROM pragma_foreign_key_list(?) WHERE "from" = ?', [table, column]))

            # We don't support compound foreign keys.
            if len(fkeys) == 1:

                other_table, other_column = fkeys[0]
                label_column = await db.label_column_for_table(other_table)
                the_other_row = list(await db.execute('SELECT "{}" FROM "{}" WHERE "{}" = ?'.format(label_column, other_table, other_column), [value]))
                if the_other_row:
                    return markupsafe.Markup(
                        '<a href="{href}">{label}</a>'.format(
                            href=datasette.urls.table(database, other_table) + '/' + str(value),
                            label=markupsafe.escape(str(the_other_row[0][0]))
                        )
                    )

        if isinstance(value, str) and (value == '[]' or (value.startswith('["') and value.endswith('"]'))):
            try:
                tags = json.loads(value)
                rv = ''

                for i, tag in enumerate(tags):
                    if i > 0:
                        rv += ', '
                    rv += markupsafe.Markup(
                        '<span>{tag}</span>'.format(
                            tag=markupsafe.escape(tag)
                        )
                    )

                return rv
            except:
                pass

    return inner

@datasette.hookimpl
def extra_body_script(template, database, table, columns, view_name, request, datasette):
    async def inner():
        facets = facets_extra_body_script(template, database, table, columns, view_name, request, datasette) or ''

        permissions = {}
        if database and table:
            for perm in ['insert-row', 'update-row', 'delete-row', 'drop-table']:
                res = await datasette.permission_allowed(
                    request.actor,
                    perm,
                    (database, table)
                )

                permissions[perm] = res == True

        if database:
            for perm in ['create-table']:
                res = await datasette.permission_allowed(
                    request.actor,
                    perm,
                    database
                )

                permissions[perm] = res == True

        permissions = '''
    __dux_permissions = {};
    '''.format(json.dumps(permissions))

        return '''
    {}

    {}
    '''.format(facets, permissions)

    return inner

@datasette.hookimpl
def startup(datasette):
    enable_yolo_facets()
    enable_yolo_arraycontains_filter()
    enable_yolo_exact_filter()
    enable_yolo_view_row_pages()
    enable_yolo_edit_row_pages()

    async def inner():
        await compute_dux_column_stats(datasette)

    return inner

@datasette.hookimpl
def register_facet_classes():
    return [StatsFacet, YearFacet, YearMonthFacet]

@datasette.hookimpl
def filters_from_request(request, database, table, datasette):
    async def dothething():
        return await yolo_filters_from_request(request, database, table, datasette)

    return dothething


@datasette.hookimpl(specname='actor_from_request', hookwrapper=True)
def sniff_actor_from_request(datasette, request):
    # TODO: This is temporary, we'll remove it when render_cell gets the request
    # param. The code is committed, just needs a new release of Datasette.
    asyncio.current_task()._duxui_request = request

    # all corresponding hookimpls are invoked here
    outcome = yield

@datasette.hookimpl
def register_routes():
    return [
        (
            css_path,
            lambda: datasette.Response(
                css_contents(),
                content_type="text/css; charset=utf-8"
            )
        ),
        (
            js_path,
            lambda: datasette.Response(
                js_contents(),
                content_type="text/javascript; charset=utf-8"
            )
        ),
        (r"^/(?P<dbname>.*)/(?P<tablename>.*)/-/dux-autosuggest-column$", handle_autosuggest_column),
        (r"^/(?P<dbname>.*)/(?P<tablename>.*)/(?P<pkey>.*)/-/dux-redirect-after-edit$", handle_redirect_after_edit)
    ]

async def handle_autosuggest_column(datasette, request):
    qs = parse_qs(request.query_string)

    column = qs['column'][0]
    q = qs.get('q', [''])[0]
    dbname = request.url_vars["dbname"]
    tablename = request.url_vars["tablename"]

    db = datasette.get_database(dbname)

    def fn(conn):
        return autosuggest_column(conn, tablename, column, q)
    suggestions = await db.execute_fn(fn)
    return Response.json(
        suggestions
    )

async def handle_redirect_after_edit(datasette, request):
    qs = parse_qs(request.query_string)

    action = qs['action'][0]
    dbname = request.url_vars["dbname"]
    tablename = request.url_vars["tablename"]
    pk = request.url_vars["pkey"]

    # TODO: why is this not finding our hook?
    redirects = pm.hook.redirect_after_edit(datasette=datasette, database=dbname, table=tablename, action=action, pk=pk)

    for redirect in redirects:
        rv = await await_me_maybe(redirect)

        if rv:
            return Response.json(
                {'url': rv}
            )

    raise Exception('handle_redirect_after_edit failed to find redirect')

@datasette.hookimpl
def get_metadata(datasette, key, database, table):
    rv = {
        'databases': {
        }
    }

    for db in datasette.databases.keys():
        # NB: don't re-use this object across databases, it causes
        # things to get mingled unexpectedly.
        hide_tables = {
            'tables': {
                DUX_IDS: { 'hidden': True },
                DUX_COLUMN_STATS: { 'hidden': True },
                DUX_COLUMN_STATS_OPS: { 'hidden': True },
                DUX_COLUMN_STATS_VALUES: { 'hidden': True },
            }
        }

        rv['databases'][db] = hide_tables

    return rv

@datasette.hookimpl
def prepare_connection(conn):
    conn.enable_load_extension(True)

    sqlite_sqlean.load(conn, 'crypto')
    conn.enable_load_extension(False)

    # Try to set synchronous = NORMAL mode
    conn.execute('PRAGMA journal_mode = WAL')
    conn.execute('PRAGMA synchronous = NORMAL')

@datasette.hookimpl
def asgi_wrapper(datasette):
    def wrap_with_indexer(app):
        @wraps(app)
        async def handle_request(scope, receive, send):
            if not hasattr(datasette, "_dux_column_stats_indexer"):
                start_column_stats_indexer(datasette)
            datasette._dux_column_stats_indexer = True
            await app(scope, receive, send)

        return handle_request

    return wrap_with_indexer



import importlib
import pluggy
import sys
from . import hookspecs

DEFAULT_PLUGINS = (
    "datasette_ui_extras.edit_controls",
)

pm = pluggy.PluginManager("datasette_ui_extras")
pm.add_hookspecs(hookspecs)

if not hasattr(sys, "_called_from_test"):
    # Only load plugins if not running tests
    pm.load_setuptools_entrypoints("datasette_ui_extras")

# Load default plugins
for plugin in DEFAULT_PLUGINS:
    mod = importlib.import_module(plugin)
    pm.register(mod, plugin)



from pluggy import HookimplMarker
from pluggy import HookspecMarker

hookspec = HookspecMarker("datasette_ui_extras")
hookimpl = HookimplMarker("datasette_ui_extras")

# Conceptually, this is a firstresult=True hook. However, we support async
# awaitables, so we need to get all the results and await them.
@hookspec()
def edit_control(datasette, database, table, column, row, value, metadata):
    """Return the name of the edit control class to use, and optionally some configuration."""

# This is also a firstresult=True hook.
@hookspec()
def redirect_after_edit(datasette, database, table, action, pk):
    """Return the URL to redirect the user to after editing a row."""



import pytest

from datasette_ui_extras.yolo_command import enable_yolo_for_metadata

def test_enable_yolo_for_metadata_empty():
    rv = enable_yolo_for_metadata({}, 'somedb')
    assert rv == {
        'databases': {
            'somedb': {
                'permissions': {
                    'create-table': True,
                    'drop-table': True,
                    'insert-row': True,
                    'update-row': True,
                    'delete-row': True,
                }
            }
        }
    }
 
def test_enable_yolo_for_metadata_existing():
    rv = enable_yolo_for_metadata({
        'databases': {
            'somedb': {
                'other': True
            }
        }
    }, 'somedb')
    assert rv == {
        'databases': {
            'somedb': {
                'other': True,
                'permissions': {
                    'create-table': True,
                    'drop-table': True,
                    'insert-row': True,
                    'update-row': True,
                    'delete-row': True,
                }
            }
        }
    }
 



from datasette import hookimpl
from datasette.filters import Filters, FilterArguments
def enable_yolo_arraycontains_filter():
    for fltr in Filters._filters:
        if fltr.key == 'arraycontains':
            fltr.sql_template = '"{c}" like \'%"\' || :{p} || \'"%\''
        if fltr.key == 'arraynotcontains':
            fltr.sql_template = '"{c}" not like \'%"\' || :{p} || \'"%\''

def patched_Filter_lookups(self):
    """Yields (lookup, display, no_argument) pairs"""
    yield 'exact', '=', False
    yield 'not', '!=', False
    for filter in self._filters:
        yield filter.key, filter.display, filter.no_argument

def enable_yolo_exact_filter():
    Filters._filters = [x for x in Filters._filters if x.key != 'exact' and x.key != 'not']
    Filters._filters_by_key = {f.key: f for f in Filters._filters}

    Filters.lookups = patched_Filter_lookups

    pass

async def yolo_filters_from_request(request, database, table, datasette):
    # Borrowed from https://github.com/simonw/datasette/blob/0b4a28691468b5c758df74fa1d72a823813c96bf/datasette/views/table.py#L347-L354
    filter_args = []
    for key in request.args:
        if not (key.startswith("_") and "__" not in key):
            for v in request.args.getlist(key):
                if not '__' in key:
                    key += '__exact'

                if key.endswith('__exact') or key.endswith('__not'):
                    filter_args.append((key, v))

    if not filter_args:
        return

    wheres = []
    params = {}
    descs = []

    param_index = 0
    for k, v in filter_args:
        op = 'exact'
        if k.endswith('__exact'):
            k = k[0:-7]
        elif k.endswith('__not'):
            op = 'not'
            k = k[0:-5]

        op_human = '=' if op == 'exact' else '!='
        wheres.append('"{c}" {op} :dux{param_index}'.format(
            c = k,
            op = op_human,
            param_index = param_index
        ))
        params['dux{}'.format(param_index)] = v

        human_desc_format = "{c} {op} {v}" if v.isdigit() else '{c} {op} "{v}"'

        human_desc = human_desc_format.format(c = k, op = op_human, v = v)

        # Is table.c an fkey to a pkey? If yes, fetch its label.
        expanded = await datasette.expand_foreign_keys(database, table, k, [v])
        if expanded:
            human_desc += ' ({})'.format(list(expanded.values())[0])
        descs.append(human_desc)
        param_index += 1
    return FilterArguments(
        wheres,
        params,
        descs
    )



import asyncio
import datasette
from datasette.utils import await_me_maybe
import markupsafe
from .hookspecs import hookimpl
import datetime
import re
import json
from .utils import row_edit_params, is_row_page
from .plugin import pm

re_iso8601_date = re.compile('^[0-9]{4}-[0-9]{2}-[0-9]{2}$')
re_iso8601_millis = re.compile('^[0-9]{4}-[0-9]{2}-[0-9]{2}([T ])[0-9]{2}:[0-9]{2}:[0-9]{2}[.][0-9]+Z?$')
re_iso8601_secs = re.compile('^[0-9]{4}-[0-9]{2}-[0-9]{2}([T ])[0-9]{2}:[0-9]{2}:[0-9]{2}Z?$')

def to_camel_case(snake_str):
    # from https://stackoverflow.com/a/19053800
    components = snake_str.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])

# From https://www.sqlite.org/datatype3.html#affinity_name_examples
numeric_types = [
    'int',
    'integer',
    'tinyint',
    'smallint',
    'mediumint',
    'bigint',
    'int2',
    'int8',
    'real',
    'double',
    'double precision',
    'float',
]


@hookimpl(trylast=True, specname='edit_control')
def string_control(datasette, database, table, column):
    return 'StringAutocompleteControl'
    #return 'StringControl'

def number_control(datasette, database, table, column, row, value, metadata):
    type = metadata['type']
    if type.lower() in numeric_types:
        return 'NumberControl'

def boolean_control(datasette, database, table, column, row, value, metadata):
    type = metadata['type']

    is_boolean = type.lower() == 'boolean'
    is_booleanish = 'min' in metadata and 'max' in metadata and isinstance(metadata['min'], int) and isinstance(metadata['max'], int) and metadata['min'] >= 0 and metadata['max'] <= 1

    if not is_boolean and not is_booleanish:
        return

    if not metadata['nullable']:
        return 'CheckboxControl'

    if type.lower() in numeric_types or type.lower() == 'boolean':
        return 'DropdownControl', { 'choices': [
            { 'value': 0, 'label': 'False' },
            { 'value': 1, 'label': 'True' },
        ]}

    return 'DropdownControl', { 'choices': [
        { 'value': '0', 'label': 'False' },
        { 'value': '1', 'label': 'True' },
    ]}

def is_iso8601_date(x):
    return re_iso8601_date.search(x)

def infer_format(metadata):
    if not 'min' in metadata or not 'max' in metadata:
        if metadata['type'].lower() == 'date':
            return {
                'precision': 'date',
                'utc': False
            }

        return

    min = metadata['min']
    max = metadata['max']

    # Maybe in the future we can support seconds since the epoch / julian days,
    # for now only ISO8601.
    if not isinstance(min, str) or not isinstance(max, str):
        return

    if re_iso8601_date.search(min) and re_iso8601_date.search(max):
        return {
            'precision': 'date',
            'utc': False,
        }

    min_z = min.endswith('Z')
    max_z = max.endswith('Z')
    min_t = 'T' in min
    max_t = 'T' in max

    min_millis = re_iso8601_millis.search(min)
    max_millis = re_iso8601_millis.search(max)

    if min_millis and max_millis and min_z == max_z and min_t == max_t:
        return {
            'precision': 'millis',
            'utc': min_z,
            't': min_t
        }

    min_secs = re_iso8601_secs.search(min)
    max_secs = re_iso8601_secs.search(max)

    if min_secs and max_secs and min_z == max_z and min_t == max_t:
        return {
            'precision': 'secs',
            'utc': min_z,
            't': min_t
        }

def date_control(datasette, database, table, column, row, value, metadata):
    # Use cases:
    # 1. type == DATE and no data: DATE
    # 2. strings that looks like date: DATE
    # 3. stringy and looks like a timestamp: whatever
    # 4. timestamptz and no data - store as ISO8601 with millis precision


    format = infer_format(metadata)

    if format:
        return 'DateControl', format

def foreign_key_control(datasette, database, table, column, row, value, metadata):
    async def inner():
        db = datasette.databases[database]
        # Is this column an fkey?
        fkeys = list(await db.execute('SELECT "table", "to" FROM pragma_foreign_key_list(?) WHERE "from" = ?', [table, column]))

        # We don't support compound foreign keys.
        if len(fkeys) == 1:
            other_table, other_column = fkeys[0]
            label_column = await db.label_column_for_table(other_table)
            the_other_row = list(await db.execute('SELECT "{}" FROM "{}" WHERE "{}" = ?'.format(label_column, other_table, other_column), [value]))
            return 'ForeignKeyControl', {
                'other_autosuggest_column_url': '{}/-/dux-autosuggest-column'.format(datasette.urls.table(database, other_table)),
                'other_table': other_table,
                'other_column': other_column,
                'label_column': label_column,
                'initial_label': the_other_row[0][0] if the_other_row else None
            }

    return inner

def textarea_control(datasette, database, table, column, row, value, metadata):
    if 'texts_newline' in metadata and metadata['texts_newline']:
        return 'TextareaControl'

def dropdown_control(datasette, database, table, column, row, value, metadata):
    if not 'check_choices' in metadata:
        return

    return 'DropdownControl', { 'choices': [{ 'value': x, 'label': x } for x in metadata['check_choices']] }

def json_tags_control(datasette, database, table, column, row, value, metadata):
    if 'json_arrays' in metadata and metadata['json_arrays'] + metadata['nulls'] == metadata['count']:
        try:
            min_parsed = json.loads(metadata['min'])
            max_parsed = json.loads(metadata['max'])

            if not isinstance(min_parsed, list) or not isinstance(max_parsed, list):
                return

            if min_parsed and not isinstance(min_parsed[0], str):
                return

            if max_parsed and not isinstance(max_parsed[0], str):
                return

            return 'JSONTagsControl'
        except:
            return

# Conceptually, it'd be nice if we just added hooks with relative priorities,
# but pluggy doesn't permit explicit priorities beyond tryfirst/trylast.
#
# Add hooks here in decreasing order of specificity
explicit_hook_order = [
    json_tags_control,
    dropdown_control,
    foreign_key_control,
    date_control,
    textarea_control,
    boolean_control,
    number_control,
    # string_control, # string_control is a trylast, so we ignore it
]

@hookimpl(specname='edit_control')
def meta_edit_control(datasette, database, table, column, row, value, metadata):
    async def inner():
        for hook in explicit_hook_order:
            rv = await await_me_maybe(hook(datasette, database, table, column, row, value, metadata))

            if rv:
                return rv

    return inner


@datasette.hookimpl(tryfirst=True, specname='render_cell')
def render_cell_edit_control(datasette, database, table, column, row, value):
    async def inner():
        task = asyncio.current_task()
        request = None if not hasattr(task, '_duxui_request') else task._duxui_request

        if not is_row_page(request):
            return

        params = await row_edit_params(datasette, request, database, table)
        if params and column in params:
            db = datasette.get_database(database)

            data = params[column]

            default_value = data['default_value']
            default_value_value = None

            if default_value:
                default_value_value = list(await db.execute("SELECT {}".format(default_value)))[0][0]

            controls = pm.hook.edit_control(datasette=datasette, database=database, table=table, column=column, row=row, value=value, metadata=data)
            control = None
            for maybe in controls:
                control = await await_me_maybe(maybe)
                if control:
                    break

            if control:
                config = {}

                if isinstance(control, tuple):
                    for k, v in control[1].items():
                        config[to_camel_case(k)] = v
                    control = control[0]

                for k, v in data.items():
                    if k == 'name':
                        k = 'column'
                    config[to_camel_case(k)] = v

                if 'base_table' in data:
                    base_table = data['base_table']
                    config['autosuggestColumnUrl'] = '{}/-/dux-autosuggest-column'.format(datasette.urls.table(database, base_table))

                config['database'] = database
                config['tableOrView'] = table

                return markupsafe.Markup(
                    '<div class="dux-edit-stub" data-control="{control}" data-initial-value="{value}" data-config="{config}">Loading...</div>'.format(
                        control=markupsafe.escape(control),
                        value=markupsafe.escape(json.dumps(value)),
                        config=markupsafe.escape(json.dumps(config)),
                    )
                )

    return inner

@hookimpl(trylast=True)
def redirect_after_edit(datasette, database, table, action, pk):
    if action == 'delete-row':
        return datasette.urls.table(database, table)
    return '{}/{}'.format(datasette.urls.table(database, table), pk)



import sqlite3

DUX_COLUMN_STATS = 'dux_column_stats'
DUX_COLUMN_STATS_OPS = 'dux_column_stats_ops'
DUX_COLUMN_STATS_VALUES = 'dux_column_stats_values'
DUX_IDS = 'dux_ids'

# This table should have reasonable defaults so we can insert new rows
# without having knowledge about what stats are being collected.
CREATE_DUX_COLUMN_STATS = '''
CREATE TABLE {}(
  table_id integer not null references dux_ids(id),
  column_id integer not null references dux_ids(id),
  type text not null,
  nullable boolean not null,
  pk boolean not null,
  min any,
  max any,
  count integer not null default 0, -- count(*)
  nulls integer not null default 0, -- the output of COUNT(*) FILTER (WHERE TYPEOF(column) == 'null')
  integers integer not null default 0, -- as above, but integer
  reals integer not null default 0, -- as above, but real
  texts integer not null default 0, -- as above, but text
  blobs integer not null default 0, -- as above, but blob
  json_strings integer not null default 0, -- sum(json_valid(column) and json_type(column) = 'text')
  json_arrays integer not null default 0, -- sum(json_valid(column) and json_type(column) = 'array')
  json_objects integer not null default 0, -- sum(json_valid(column) and json_type(column) = 'object')
  texts_min_length integer default 0,
  texts_max_length integer default 0,
  texts_whitespace integer not null default 0, -- # of texts that have whitespace
  texts_newline integer not null default 0, -- # of texts that have a newline
  blobs_min_length integer default 0,
  blobs_max_length integer default 0,
  primary key (table_id, column_id)
)
'''.format(DUX_COLUMN_STATS).strip()

CREATE_DUX_COLUMN_STATS_VALUES = '''
CREATE TABLE {}(
  table_id integer not null references dux_ids(id),
  column_id integer not null references dux_ids(id),
  value text,
  hash blob not null,
  count integer not null,
  pks text not null, -- JSON array of pkeys that have this value
  primary key (table_id, column_id, value, hash)
)
'''.format(DUX_COLUMN_STATS_VALUES).strip()

CREATE_DUX_IDS = '''
CREATE TABLE {}(
  id integer primary key,
  name text not null unique
)
'''.format(DUX_IDS).strip()

CREATE_DUX_COLUMN_STATS_OPS = '''
CREATE TABLE {}(
  table_id integer REFERENCES dux_ids(id),
  column_id integer REFERENCES dux_ids(id),
  last_key text not null default '{{}}',
  pending integer not null check (pending in (0, 1)),
  updated_at text not null default '1970-01-01 00:00:00Z',
  primary key (table_id, column_id)
)
'''.format(DUX_COLUMN_STATS_OPS).strip()

table_schemas = {
    DUX_IDS: CREATE_DUX_IDS,
    DUX_COLUMN_STATS_OPS: CREATE_DUX_COLUMN_STATS_OPS,
    DUX_COLUMN_STATS: CREATE_DUX_COLUMN_STATS,
    DUX_COLUMN_STATS_VALUES: CREATE_DUX_COLUMN_STATS_VALUES
}

async def ensure_schema(db):
    # We take a very brute force approach: if any table has the wrong schema,
    # we'll drop all the tables.
    await db.execute_write_fn(ensure_schema_conn)

def ensure_schema_conn(conn):
    def get_sql(table):
        sql = conn.execute("SELECT sql FROM sqlite_master WHERE name = ?", [table]).fetchone()

        if not sql:
            return None

        return sql[0]

    all_ok = True
    for k, v in table_schemas.items():
        if get_sql(k) != v:
            all_ok = False
            break

    if all_ok:
        return

    # Drop all the tables, in reverse order so fkeys don't cause us grief.
    for table in reversed(list(table_schemas.keys())):
        conn.execute('DROP TABLE IF EXISTS {}'.format(table))

    for table, sql in table_schemas.items():
        conn.execute(sql)

        roundtrip = get_sql(table)
        if roundtrip != sql:
            raise Exception('failed to create sql for table {}\nexpected: {}\nactual: {}'.format(table, sql, roundtrip))




from urllib.parse import urlparse, parse_qs
from sqlglot import parse_one, exp
from .column_stats_schema import DUX_COLUMN_STATS, DUX_IDS
from .schema_utils import get_column_choices_from_check_constraints

def is_row_page(request):
    # This feels like a giant hack!
    return request and request.scope and request.scope['url_route'] and 'database' in request.scope['url_route']['kwargs'] and 'table' in request.scope['url_route']['kwargs'] and 'pks' in request.scope['url_route']['kwargs'] and True

# Returns:
# - None if the edit UI ought not be shown
# - Information about which columns are editable
async def row_edit_params(datasette, request, database, table):
    if not request:
        return None

    if hasattr(request, '_dux_row_edit_params'):
        return request._dux_row_edit_params

    rv = await row_edit_params_raw(datasette, request, database, table)
    request._dux_row_edit_params = rv
    return rv

async def row_edit_params_raw(datasette, request, database, table):
    url = urlparse(request.url)
    qs = parse_qs(url.query)

    dux_edit_param = '_dux_edit' in qs and qs['_dux_edit'] == ['1']

    config = datasette.plugin_config('datasette-ui-extras', database=database)

    metadata_edit_param = config and 'editable' in config and table in config['editable']

    if not (dux_edit_param or metadata_edit_param):
        return None

    # Ensure user has permission to update this row
    allowed = await datasette.permission_allowed(
        request.actor,
        "update-row",
        (database, table)
    )

    if not allowed:
        return None

    rv = await get_editable_columns(datasette, request, database, table)
    return rv

async def get_editable_columns(datasette, request, database, table):
    db = datasette.get_database(database)

    is_view = await db.view_exists(table)


    pkeys = await db.primary_keys(table)
    columns = await db.table_columns(table)


    if not is_view:
        def fn(conn):
            return get_table_info(conn, table)

        table_info = await db.execute_fn(fn)

        rv = {}
        for column in table_info['columns'].values():
            if column['pk']:
                continue

            rv[column['name']] = column

        await annotate_columns(rv, db, table)

        return rv

    def fn(conn):
        return get_view_info(conn, table)

    view_info = await db.execute_fn(fn)

    if not view_info:
        return {}

    rv = {}
    for column in view_info['base_columns']:
        if column in view_info['pks']:
            continue
        rv[column] = view_info['table_info']['columns'][column]

    await annotate_columns(rv, db, view_info['base_table'])

    return rv

async def annotate_columns(rv, db, table_name):
    # Annotate rv with data from a few places:
    # (1) sqlite_master - parse the table schema for CHECK (column IN ('a', 'b', ..., 'z')) constraints
    #     These are a closed set of options.
    #
    # (2) dux_column_stats - this is summary metadata that can inform what UI is best to show.
    schema = list(await db.execute("SELECT sql FROM sqlite_master WHERE name = ? AND type = 'table'", [table_name]))[0][0]
    choices = get_column_choices_from_check_constraints(schema)
    for column, options in choices.items():
        if column in rv:
            rv[column]['check_choices'] = options


    results = []
    try:
        # TODO: this will need to look up by dux_ids
        results = list(await db.execute('SELECT stats.*, ids.name AS column FROM dux_column_stats stats JOIN dux_ids ids ON ids.id = stats.column_id WHERE "table_id" = (SELECT id FROM dux_ids WHERE name = ?)', [table_name]))
    except:
        return

    for row in results:
        column = row['column']
        if not column in rv:
            continue

        to_annotate = rv[column]
        to_annotate['base_table'] = table_name

        for key in row.keys():
            if key == 'column':
                continue
            elif key == 'table_id' or key == 'column_id':
                # Don't pass these through -- they're internal transient details, as dux_ids IDs aren't
                # meaningful
                pass
            else:
                value = row[key]
                if key == 'nullable':
                    value = not not value
                to_annotate[key] = value


def get_table_info(conn, name):
    data = conn.execute('select name, "type", "notnull", dflt_value, pk from pragma_table_info(?)', [name]).fetchall()
    rv = {}
    cols = {}
    rv['columns'] = cols

    # TODO: add sqlglot to parse CHECK constraints

    # TODO: parse select "table", "from", "to" from pragma_foreign_key_list(?)
    for name, type, notnull, dflt_value, pk in data:
        cols[name] = {
            'name': name,
            'type': type,
            'nullable': notnull == 0,
            'default_value': dflt_value,
            'pk': pk == 1
        }

    return rv

def get_view_info(conn, name):
    sql, = conn.execute("select sql from sqlite_master where type = 'view' and name = ?", [name]).fetchone()

    parsed = None
    try:
        parsed = parse_one(sql)
    except:
        return {}

    view_exp = parsed.expression

    if not 'from' in view_exp.args:
        return {}

    from_exps = view_exp.args['from'].expressions

    if len(from_exps) != 1:
        return {}

    table_name = from_exps[0].this.this

    table_pkeys = [x[0] for x in conn.execute('select name from pragma_table_info(?) where pk = 1', [table_name]).fetchall()]
    cols = view_exp.expressions

    # We only want the columns that are pass-through identifiers
    ids = []
    for col in cols:
        this = col.this
        if isinstance(col, exp.Column) and isinstance(col.this, exp.Identifier):
            ids.append(col.name)


    return {
        'view': name,
        'base_table': table_name,
        'pks': table_pkeys,
        'base_columns': ids,
        'table_info': get_table_info(conn, table_name)
    }





from datasette import hookimpl
import click
import sqlite3
import json
import os

@hookimpl(specname='register_commands')
def yolo_command(cli):
    @cli.command()
    @click.argument(
        "files", type=click.Path(), nargs=-1
    )
    def yolo(files):
        "Enable very permissive access to the specified databases."

        # First, enable WAL mode for all the databases, creating them
        # if needed.
        for file in files:
            conn = sqlite3.connect(str(file))
            try:
                conn.execute('PRAGMA journal_mode = WAL')
            except sqlite3.DatabaseError:
                raise click.ClickException(
                    "Invalid database: {}".format(file)
                )

        # Second, ensure there is a metadata.json that is permissive
        metadata = {}
        indent = 4
        try:
            f = open('metadata.json', 'r')
            raw_json = f.read()
            metadata = json.loads(raw_json)
            if not '    ' in raw_json:
                indent = 2
            f.close()
        except FileNotFoundError:
            pass

        for file in files:
            db = os.path.basename(file)
            if '.' in db:
                db = '.'.join(db.split('.')[0:-1])
            metadata = enable_yolo_for_metadata(metadata, db)


        f = open('metadata.json', 'w')
        f.write(json.dumps(metadata, indent=indent))
        f.close()

def enable_yolo_for_metadata(metadata, db):
    rv = json.loads(json.dumps(metadata))

    if not 'databases' in rv:
        rv['databases'] = {}

    if not db in rv['databases']:
        rv['databases'][db] = {}

    if not 'permissions' in rv['databases'][db]:
        rv['databases'][db]['permissions'] = {}

    for perm in ['create-table', 'drop-table', 'insert-row', 'update-row', 'delete-row']:
        rv['databases'][db]['permissions'][perm] = True

    return rv



from sqlglot import parse_one, exp

def get_column_choices_from_check_constraints(sql):
    if not isinstance(sql, str):
        raise Exception('expected sql to be str but was {}'.format(sql))

    try:
        parsed = parse_one(sql)
    except:
        # They might have used something unparseable by sqlglot, fail gracefully.
        return {}

    # Returns a map from column name to permitted values, if
    # table defn has column defn of the form
    # x check (x in (...))
    rv = {}
    checks = list(parsed.find_all(exp.CheckColumnConstraint, exp.Check))

    for check in checks:
        if not isinstance(check.this, exp.In):
            continue

        in_ = check.this
        exprs = in_.expressions
        column = in_.this.this.this
        ok = []

        # Are all of the exprs literals?
        all_literal = True
        for expr in exprs:
            if not isinstance(expr, exp.Literal):
                all_literal = False
                break

            if expr.is_string:
                ok.append(expr.this)
            else:
                ok.append(int(expr.this))


        if not all_literal:
            continue

        if ok:
            rv[column] = ok

    return rv
